@@ -5,6 +5,20 @@ from ABlooper.models import DecoyGen
 from ABlooper.utils import filt, rmsd, to_pdb_line, short2long, long2short, prepare_input_loop
 import os
 
+try:
+    from ABlooper.rosetta_refine import rosetta_refine
+except ModuleNotFoundError:
+    rosetta_available = False
+else:
+    rosetta_available = True
+
+try:
+    from ABlooper.openmm_refine import openmm_refine
+except ModuleNotFoundError:
+    openmm_available = False
+else:
+    openmm_available = True
+
 device = "cuda" if torch.cuda.is_available() else "cpu"
 current_directory = os.path.dirname(os.path.realpath(__file__))
 
@@ -14,7 +28,7 @@ default_model.load_state_dict(torch.load(path, map_location=torch.device(device)
 
 
 class CDR_Predictor:
-    def __init__(self, pdb_file, chains=("H", "L"), model=None):
+    def __init__(self, pdb_file, chains=("H", "L"), model=None, refine=False, refine_method="openmm"):
         """ Class used to handle remodelling of CDRs.
 
         :param pdb_file: File of IMGT numbered antibody structure in .pdb format. Must include heavy and light chain
@@ -54,6 +68,12 @@ class CDR_Predictor:
         self.CDR_start_atom_id = {CDR: int([x.split()[1] for x in self.CDR_text[CDR] if x.split()[2] == "N"][2]) for CDR
                                   in self.CDR_text}
 
+        # If rosetta or openmm are not available refinement is not possible
+        if (not rosetta_available) and (not openmm_available) and refine:
+            print("Neither PyRosetta nor OpenMM are available. Refinement is not possible")
+        self.refine = refine and (rosetta_available or openmm_available)
+        self.refine_method = refine_method
+
     def __extract_BB_coords(self):
         self.CDR_BB_coords = {}
 
@@ -73,7 +93,11 @@ class CDR_Predictor:
                     i += 1
                 if cut[2] in self.__atoms:
                     j = self.__atoms.index(cut[2])
-                    coors[i, j] = np.float_(cut[6:9])
+                    # Using split for corrds doesn't always work. Following Biopython approach:
+                    x = float(line[30:38])
+                    y = float(line[38:46])
+                    z = float(line[46:54])
+                    coors[i, j] = np.array([x, y, z])
 
             # If missed CB (GLY) then add CA instead
             coors[:, 3] = np.where(np.all(coors[:, 3] != coors[:, 3], axis=-1, keepdims=True), coors[:, 0], coors[:, 3])
@@ -165,7 +189,18 @@ class CDR_Predictor:
                     continue
             old_text = new_text
 
-        new_text = "REMARK   REMODELLED CDRs USING ABlooper \n" + "".join(old_text)
+        header = [
+            "REMARK    CDR LOOPS REMODELLED USING ABLOOPER                                   \n"]
+
+        if self.refine:
+            if self.refine_method == "openmm" and openmm_available:
+                old_text = openmm_refine(old_text, self.CDR_with_anchor_slices)
+                header.append("REMARK    REFINEMENT DONE USING OPENMM" + 42 * " " + "\n")
+            else:
+                old_text = rosetta_refine(old_text, self.CDR_with_anchor_slices)
+                header.append("REMARK    REFINEMENT DONE USING PYROSETTA" + 39 * " " + "\n")
+
+        new_text = "".join(header + old_text)
 
         if file_name is None:
             return "".join(new_text)
